from __future__ import annotations
import numpy as np


def init_population(pop_size: int) -> np.ndarray:
    return np.random.randint(0, 8, (pop_size, 8))


def fitness(population: np.ndarray) -> np.ndarray:
    fitness_vals = []
    for individual in population:
        conflicts = 0
        for i in range(len(individual)):
            row = int(individual[i])
            for j in range(len(individual)):
                if i == j:
                    continue
                d = abs(i - j)
                if individual[j] in [row, row - d, row + d]:
                    conflicts += 1
        fitness_vals.append(conflicts)
    return -1 * np.array(fitness_vals, dtype=float)


def selection(population: np.ndarray, fitness_vals: np.ndarray) -> np.ndarray:
    probs = fitness_vals.copy()
    min_fitness = abs(probs.min()) + 1.0
    probs = probs + min_fitness
    probs = probs / probs.sum()

    n = len(population)
    indices = np.arange(n)
    selected_indices = np.random.choice(indices, size=n, p=probs)
    return population[selected_indices]


def crossover(parent1: np.ndarray, parent2: np.ndarray, pc: float) -> tuple[np.ndarray, np.ndarray]:
    r = np.random.random()
    if r < pc:
        point = np.random.randint(1, 8)
        child1 = np.concatenate([parent1[:point], parent2[point:]])
        child2 = np.concatenate([parent2[:point], parent1[point:]])
    else:
        child1 = parent1.copy()
        child2 = parent2.copy()
    return child1, child2


def mutation(individual: np.ndarray, pm: float) -> np.ndarray:
    r = np.random.random()
    if r < pm:
        m = np.random.randint(8)
        individual[m] = np.random.randint(8)
    return individual


def crossover_mutation(selected_pop: np.ndarray, pc: float, pm: float) -> np.ndarray:
    n = len(selected_pop)
    new_pop = np.empty((n, 8), dtype=int)

    for i in range(0, n, 2):
        parent1 = selected_pop[i]
        parent2 = selected_pop[i + 1]
        child1, child2 = crossover(parent1, parent2, pc)
        new_pop[i] = child1
        new_pop[i + 1] = child2

    for i in range(n):
        new_pop[i] = mutation(new_pop[i], pm)
    return new_pop


def run_eight_queens_ga(pop_size: int = 100, max_generations: int = 1000, pc: float = 0.7, pm: float = 0.01, seed: int | None = None) -> dict:
    if seed is not None:
        np.random.seed(seed)

    population = init_population(pop_size)
    best_fitness_overall = None
    best_solution = None
    best_gen = None

    for gen in range(max_generations):
        fitness_vals = fitness(population)
        best_i = int(fitness_vals.argmax())
        best_fit = float(fitness_vals[best_i])

        if best_fitness_overall is None or best_fit > best_fitness_overall:
            best_fitness_overall = best_fit
            best_solution = population[best_i].copy()
            best_gen = gen

        if best_fitness_overall == 0:
            break

        selected_pop = selection(population, fitness_vals)
        if len(selected_pop) % 2 == 1:
            selected_pop = selected_pop[:-1]
        population = crossover_mutation(selected_pop, pc, pm)

    return {
        "best_solution": best_solution,
        "best_fitness": best_fitness_overall,
        "generation": best_gen,
    }



import requests
from typing import List

URL = "https://mercury-coder-ai.p.rapidapi.com/chat/completions"

API_KEYS: List[str] = [
    "d79ff173d3msh0401728dbbf5937p1e3d8ejsnc5e984dcdc17",
    "177f77e0e3msh3a667086b70240ap1bb9c2jsn70991ea865e8",
    "27e6e0feb3mshe575bbbf152aa63p164046jsne6db7318a2b6",
    "6d0546729fmsh01d7676cf9e4be2p13caa2jsn29e0bcff1606",
    "3f86b5808cmsh539f9359c9e45acp1bebcdjsnc3cfa2279a59",
    "56bbedf133mshc1278641abcbae0p16156djsn11f09e3e5a3f",
    "562dfece47mshbdd3c422ec757cfp118cb7jsn8f6106a399e2",
    "48890f6a43msh245e941d6e13592p1776a7jsne8500fc63692",
    "5a377e926emshf095eacc90d9c16p1d2ba0jsn3ba234aadc7b",
    "3857fe4f62mshf797cfd30db2980p11345bjsn2ea58a083625",
    "1f1a55f1e3msh111301c22fd6abcp11e536jsn1f7ca68b3f1f",
    "66a90182a0mshdecfd173328829ep13fa7bjsnabe48bf0e5fd",
    "f2491cba4cmsh8761608d604a8aap1e0410jsndcb59880b6df"
]

SYSTEM_PROMPT = """
You are a coding assistant specialized in Genetic algorithms. Rules:
1. Return ONLY valid Python code.
2. Do not use Markdown formatting (no ```python).
3. Do not include explanations.
4. Import all necessary libraries (especially numpy).
5. The solution must use a Genetic Algorithm approach.
"""

def fitness(user_prompt: str, output_file: str = "solution.py") -> str:
    payload = {
        "model": "mercury-coder",
        "messages": [
            {"role": "system", "content": SYSTEM_PROMPT},
            {"role": "user", "content": user_prompt},
        ],
    }

    last_error = None

    for key in API_KEYS:
        headers = {
            "x-rapidapi-key": key,
            "x-rapidapi-host": "mercury-coder-ai.p.rapidapi.com",
            "Content-Type": "application/json",
        }

        try:
            response = requests.post(URL, json=payload, headers=headers, timeout=2000)
            response.raise_for_status()

            data = response.json()
            content = data["choices"][0]["message"]["content"]

            with open(output_file, "w", encoding="utf-8") as f:
                f.write(content)

            print(content)
            return content

        except Exception as e:
            last_error = e
            continue

    raise RuntimeError(f"All API keys failed. Last error: {last_error}")



from __future__ import annotations
import argparse

from .linear_equation_ga import run_linear_equation_ga
from .tsp_ga import run_tsp_ga
from .eight_queens_ga import run_eight_queens_ga
from .graph_coloring_ga import run_graph_coloring_ga
from .continuous_opt_ga import run_continuous_opt_ga
from .knapsack_ga import run_knapsack_ga
from .symbolic_gp import run_symbolic_gp
from .runner import fitness




from __future__ import annotations
import argparse

from .linear_equation_ga import run_linear_equation_ga
from .tsp_ga import run_tsp_ga
from .eight_queens_ga import run_eight_queens_ga
from .graph_coloring_ga import run_graph_coloring_ga
from .continuous_opt_ga import run_continuous_opt_ga
from .knapsack_ga import run_knapsack_ga
from .symbolic_gp import run_symbolic_gp


def main() -> None:
    parser = argparse.ArgumentParser(prog="genetic-toolkit")
    sub = parser.add_subparsers(dest="cmd", required=True)

    sub.add_parser("linear")
    tsp = sub.add_parser("tsp")
    tsp.add_argument("--cities", type=int, default=10)
    tsp.add_argument("--gens", type=int, default=100)

    sub.add_parser("queens")
    sub.add_parser("coloring")
    sub.add_parser("continuous")
    sub.add_parser("knapsack")
    sub.add_parser("gp")

    parser.add_argument("--seed", type=int, default=None)

    args = parser.parse_args()

    if args.cmd == "linear":
        print(run_linear_equation_ga(seed=args.seed))
    elif args.cmd == "tsp":
        print(run_tsp_ga(n_cities=args.cities, generations=args.gens, seed=args.seed))
    elif args.cmd == "queens":
        print(run_eight_queens_ga(seed=args.seed))
    elif args.cmd == "coloring":
        print(run_graph_coloring_ga(seed=args.seed))
    elif args.cmd == "continuous":
        print(run_continuous_opt_ga(seed=args.seed))
    elif args.cmd == "knapsack":
        print(run_knapsack_ga(seed=args.seed))
    elif args.cmd == "gp":
        print(run_symbolic_gp(seed=args.seed))



from __future__ import annotations
import numpy as np


def initialize_population(n: int, gene_min: int = 0, gene_max: int = 30) -> np.ndarray:
    return np.random.randint(gene_min, gene_max, (n, 4))


def compute_fitness(chromosome: np.ndarray) -> tuple[np.ndarray, np.ndarray, float, np.ndarray, np.ndarray]:
    objective = np.abs(30 - (chromosome[:, 0] + 2 * chromosome[:, 1] + 3 * chromosome[:, 2] + 4 * chromosome[:, 3]))
    fitness = 1.0 / (1.0 + objective)

    total = float(fitness.sum())
    prob = fitness / total
    cum_sum = np.cumsum(prob)
    return objective, fitness, total, prob, cum_sum


def selection(chromosome: np.ndarray, cum_sum: np.ndarray) -> np.ndarray:
    rand_nums = np.random.random(chromosome.shape[0])
    new_chromosome = np.zeros((chromosome.shape[0], 4), dtype=int)

    for i in range(len(rand_nums)):
        for j in range(len(cum_sum)):
            if rand_nums[i] < cum_sum[j]:
                new_chromosome[i, :] = chromosome[j, :]
                break
    return new_chromosome


def crossover(chromosome: np.ndarray, pc: float = 0.25) -> np.ndarray:
    n = chromosome.shape[0]
    rand_nums = np.random.random(n)
    flag = rand_nums < pc

    cross_chromosome = chromosome[flag]
    len_cross = len(cross_chromosome)
    if len_cross == 0:
        return chromosome

    cross_values = np.random.randint(1, 3, len_cross)
    cpy_chromosome = np.copy(cross_chromosome)

    for i in range(len_cross):
        c_val = int(cross_values[i])
        if i == len_cross - 1:
            cross_chromosome[i, c_val:] = cpy_chromosome[0, c_val:]
        else:
            cross_chromosome[i, c_val:] = cpy_chromosome[i + 1, c_val:]

    idx = 0
    for i in range(n):
        if flag[i]:
            chromosome[i, :] = cross_chromosome[idx, :]
            idx += 1
    return chromosome


def mutation(chromosome: np.ndarray, pm: float = 0.1, gene_min: int = 0, gene_max: int = 30) -> np.ndarray:
    total_gen = chromosome.size
    no_of_mutations = int(np.round(pm * total_gen))
    if no_of_mutations <= 0:
        return chromosome

    gen_num = np.random.randint(0, total_gen - 1, no_of_mutations)
    replacing_num = np.random.randint(gene_min, gene_max, no_of_mutations)

    for i in range(no_of_mutations):
        a = int(gen_num[i])
        row = a // 4
        col = a % 4
        chromosome[row, col] = int(replacing_num[i])
    return chromosome


def run_linear_equation_ga(n: int = 6, epochs: int = 50, seed: int | None = None) -> dict:
    if seed is not None:
        np.random.seed(seed)

    chromosome = initialize_population(n)

    for epoch in range(epochs):
        objective, fitness_vals, total, prob, cum_sum = compute_fitness(chromosome)
        best_idx = int(np.argmax(fitness_vals))

        if int(objective[best_idx]) == 0:
            return {
                "found": True,
                "epoch": epoch,
                "best": chromosome[best_idx].copy(),
                "objective": int(objective[best_idx]),
                "fitness": float(fitness_vals[best_idx]),
                "final_population": chromosome.copy(),
            }

        chromosome = selection(chromosome, cum_sum)
        chromosome = crossover(chromosome)
        chromosome = mutation(chromosome)

    objective, fitness_vals, *_ = compute_fitness(chromosome)
    best_idx = int(np.argmax(fitness_vals))
    return {
        "found": False,
        "epoch": epochs,
        "best": chromosome[best_idx].copy(),
        "objective": int(objective[best_idx]),
        "fitness": float(fitness_vals[best_idx]),
        "final_population": chromosome.copy(),
    }



from __future__ import annotations
import random
import numpy as np


def create_initial_population(pop_size: int) -> list[float]:
    return [random.uniform(0, 2) for _ in range(pop_size)]


def fitness_function(x: float) -> float:
    return float(x * np.sin(10 * x) + x)


def tournament_selection(pop: list[float], k: int = 3) -> float:
    selected = random.sample(pop, k)
    selected.sort(key=fitness_function, reverse=True)
    return selected[0]


def crossover_uniform(parent1: float, parent2: float, swap_rate: float = 0.5) -> float:
    if random.random() < swap_rate:
        alpha = random.random()
        return float(alpha * parent1 + (1 - alpha) * parent2)
    return float(parent1)


def mutate(child: float, mutation_rate: float = 0.1) -> float:
    if random.random() < mutation_rate:
        child += random.uniform(-0.2, 0.2)
        child = max(0.0, min(2.0, child))
    return float(child)


def run_continuous_opt_ga(generations: int = 400, pop_size: int = 100, seed: int | None = None) -> dict:
    if seed is not None:
        random.seed(seed)
        np.random.seed(seed)

    population = create_initial_population(pop_size)

    for _ in range(generations):
        new_population = []
        for _ in range(pop_size):
            parent1 = tournament_selection(population)
            parent2 = tournament_selection(population)
            child = crossover_uniform(parent1, parent2)
            child = mutate(child)
            new_population.append(child)
        population = new_population

    best = max(population, key=fitness_function)
    return {"best_x": best, "best_value": fitness_function(best)}



from __future__ import annotations
import random
import numpy as np


def create_route(n: int) -> list[int]:
    route = list(range(n))
    random.shuffle(route)
    return route


def initial_population(pop_size: int, n: int) -> list[list[int]]:
    return [create_route(n) for _ in range(pop_size)]


def distance(a: tuple[float, float], b: tuple[float, float]) -> float:
    return float(np.sqrt((a[0] - b[0]) ** 2 + (a[1] - b[1]) ** 2))


def route_distance(route: list[int], cities: list[tuple[float, float]]) -> float:
    total = 0.0
    for i in range(len(route) - 1):
        total += distance(cities[route[i]], cities[route[i + 1]])
    total += distance(cities[route[-1]], cities[route[0]])
    return float(total)


def fitness(route: list[int], cities: list[tuple[float, float]]) -> float:
    d = route_distance(route, cities)
    return 1.0 / d if d > 0 else float("inf")


def crossover(parent_1: list[int], parent_2: list[int], n_cities: int) -> tuple[list[int], list[int]]:
    cut = random.randint(1, n_cities - 1)
    offspring_1 = parent_1[0:cut] + [city for city in parent_2 if city not in parent_1[0:cut]]
    offspring_2 = parent_2[0:cut] + [city for city in parent_1 if city not in parent_2[0:cut]]
    return offspring_1, offspring_2


def mutate(route: list[int], mutation_rate: float) -> list[int]:
    for i in range(len(route)):
        if random.random() < mutation_rate:
            j = random.randint(0, len(route) - 1)
            route[i], route[j] = route[j], route[i]
    return route


def genetic_algorithm(
    cities: list[tuple[float, float]],
    pop_size: int = 100,
    top_parents: int = 20,
    mutation_rate: float = 0.01,
    generations: int = 300,
    seed: int | None = None,
) -> tuple[list[tuple[list[int], int, float]], list[int], float]:
    if seed is not None:
        random.seed(seed)
        np.random.seed(seed)

    n = len(cities)
    population = initial_population(pop_size, n)
    history: list[tuple[list[int], int, float]] = []

    for gen in range(generations):
        population = sorted(population, key=lambda x: fitness(x, cities), reverse=True)
        parents = population[:top_parents]

        children: list[list[int]] = []
        while len(children) < (pop_size - top_parents):
            parent1, parent2 = random.sample(parents, 2)
            child1, child2 = crossover(parent1, parent2, n)
            children.append(mutate(child1, mutation_rate))
            if len(children) < (pop_size - top_parents):
                children.append(mutate(child2, mutation_rate))

        population = parents + children
        best_route = population[0]
        best_dist = route_distance(best_route, cities)
        history.append((best_route[:], gen, best_dist))

    best_route = sorted(population, key=lambda x: fitness(x, cities), reverse=True)[0]
    best_distance = route_distance(best_route, cities)
    return history, best_route, best_distance


def run_tsp_ga(
    n_cities: int = 10,
    pop_size: int = 50,
    top_parents: int = 10,
    mutation_rate: float = 0.01,
    generations: int = 100,
    seed: int | None = None,
) -> dict:
    if seed is not None:
        random.seed(seed)
        np.random.seed(seed)

    cities = [(random.uniform(0, 100), random.uniform(0, 100)) for _ in range(n_cities)]
    history, best_route, best_distance = genetic_algorithm(
        cities,
        pop_size=pop_size,
        top_parents=top_parents,
        mutation_rate=mutation_rate,
        generations=generations,
        seed=seed,
    )
    return {
        "cities": cities,
        "history": history,
        "best_route": best_route,
        "best_distance": best_distance,
    }



from __future__ import annotations
import random


DEFAULT_VALUES = [60, 100, 120, 80, 150]
DEFAULT_WEIGHTS = [10, 20, 30, 25, 50]
DEFAULT_MAX_WEIGHT = 50


def initialize_population(pop_size: int, num_items: int) -> list[list[int]]:
    return [[random.randint(0, 1) for _ in range(num_items)] for _ in range(pop_size)]


def fitness(chromosome: list[int], values: list[int], weights: list[int], max_weight: int) -> int:
    total_value = sum(chromosome[i] * values[i] for i in range(len(chromosome)))
    total_weight = sum(chromosome[i] * weights[i] for i in range(len(chromosome)))
    if total_weight > max_weight:
        return 0
    return total_value


def selection(population: list[list[int]], values: list[int], weights: list[int], max_weight: int) -> list[list[int]]:
    population = sorted(population, key=lambda ch: fitness(ch, values, weights, max_weight), reverse=True)
    return population[: len(population) // 2]


def crossover(parent1: list[int], parent2: list[int]) -> tuple[list[int], list[int]]:
    if len(parent1) > 1:
        point = random.randint(1, len(parent1) - 1)
        child1 = parent1[:point] + parent2[point:]
        child2 = parent2[:point] + parent1[point:]
        return child1, child2
    return parent1[:], parent2[:]


def mutate(chromosome: list[int], mutation_rate: float = 0.1) -> list[int]:
    for i in range(len(chromosome)):
        if random.random() < mutation_rate:
            chromosome[i] = 1 - chromosome[i]
    return chromosome


def run_knapsack_ga(
    pop_size: int = 10,
    generations: int = 50,
    mutation_rate: float = 0.1,
    values: list[int] | None = None,
    weights: list[int] | None = None,
    max_weight: int = DEFAULT_MAX_WEIGHT,
    seed: int | None = None,
) -> dict:
    if seed is not None:
        random.seed(seed)

    values = values or DEFAULT_VALUES
    weights = weights or DEFAULT_WEIGHTS

    population = initialize_population(pop_size, len(values))

    for _ in range(generations):
        selected = selection(population, values, weights, max_weight)
        next_generation: list[list[int]] = []

        while len(next_generation) < pop_size:
            parent1, parent2 = random.sample(selected, 2)
            child1, child2 = crossover(parent1, parent2)
            next_generation.append(mutate(child1, mutation_rate))
            if len(next_generation) < pop_size:
                next_generation.append(mutate(child2, mutation_rate))

        population = next_generation

    best_solution = max(population, key=lambda ch: fitness(ch, values, weights, max_weight))
    best_value = fitness(best_solution, values, weights, max_weight)
    best_weight = sum(best_solution[i] * weights[i] for i in range(len(best_solution)))

    return {"best_items": best_solution, "best_value": best_value, "best_weight": best_weight}



from __future__ import annotations
import random
from typing import Any


def generate_random_tree(depth: int = 3) -> Any:
    if depth == 0:
        return random.choice(["x", "y", random.randint(1, 10)])
    operator = random.choice(["+", "-", "*", "/"])
    return [operator, generate_random_tree(depth - 1), generate_random_tree(depth - 1)]


def evaluate(tree: Any, vars: dict[str, float]) -> float:
    if isinstance(tree, (int, float)):
        return float(tree)
    if isinstance(tree, str):
        return float(vars[tree])

    operator, left, right = tree
    left_val = evaluate(left, vars)
    right_val = evaluate(right, vars)

    if operator == "+":
        return left_val + right_val
    if operator == "-":
        return left_val - right_val
    if operator == "*":
        return left_val * right_val
    if operator == "/":
        return left_val / right_val if right_val != 0 else float("inf")

    raise ValueError(f"Unknown operator: {operator}")


def fitness(tree: Any, target: float, vars: dict[str, float]) -> float:
    try:
        result = evaluate(tree, vars)
        return -abs(result - target)
    except Exception:
        return float("-inf")


def selection_gp(population: list[Any], fitness_scores: list[float]) -> list[Any]:
    selected = []
    for _ in range(len(population) // 2):
        i, j = random.sample(range(len(population)), 2)
        selected.append(population[i] if fitness_scores[i] > fitness_scores[j] else population[j])
    return selected


def crossover_gp(parent1: Any, parent2: Any) -> Any:
    if not isinstance(parent1, list) or not isinstance(parent2, list):
        return parent1 if random.random() < 0.5 else parent2

    operator = parent1[0]
    if random.random() < 0.5:
        return [operator, parent2[1], parent1[2]]
    return [operator, parent1[1], parent2[2]]


def mutate_gp(tree: Any, mutation_rate: float = 0.1) -> Any:
    if random.random() < mutation_rate:
        return generate_random_tree(depth=2)
    if isinstance(tree, list):
        return [tree[0], mutate_gp(tree[1], mutation_rate), mutate_gp(tree[2], mutation_rate)]
    return tree


def run_symbolic_gp(
    variables: dict[str, float] | None = None,
    target_value: float = 21.0,
    generations: int = 50,
    pop_size: int = 20,
    mutation_rate: float = 0.1,
    seed: int | None = None,
) -> dict:
    if seed is not None:
        random.seed(seed)

    variables = variables or {"x": 4.0, "y": 5.0}

    population = [generate_random_tree(depth=3) for _ in range(pop_size)]
    best_tree = population[0]
    best_fit = float("-inf")

    for gen in range(generations):
        fitness_scores = [fitness(tree, target_value, variables) for tree in population]

        zipped = sorted(zip(fitness_scores, population), key=lambda x: x[0], reverse=True)
        population = [x[1] for x in zipped[: pop_size // 2]]

        best_tree = population[0]
        best_fit = fitness(best_tree, target_value, variables)
        if best_fit == 0:
            return {
                "found": True,
                "generation": gen,
                "best_tree": best_tree,
                "fitness": best_fit,
                "value": evaluate(best_tree, variables),
            }

        new_population = []
        while len(new_population) < pop_size:
            if len(population) < 2:
                break
            parent1, parent2 = random.sample(population, 2)
            child = crossover_gp(parent1, parent2)
            child = mutate_gp(child, mutation_rate)
            new_population.append(child)

        population = new_population[:pop_size]

    return {
        "found": False,
        "generation": generations,
        "best_tree": best_tree,
        "fitness": best_fit,
        "value": evaluate(best_tree, variables),
    }



from __future__ import annotations
import random


def random_chromosome(num_genes: int, num_colors: int) -> list[int]:
    return [random.randint(0, num_colors - 1) for _ in range(num_genes)]


def initial_population(pop_size: int, num_genes: int, num_colors: int) -> list[list[int]]:
    return [random_chromosome(num_genes, num_colors) for _ in range(pop_size)]


def fitness(chromosome: list[int], graph: dict[int, list[int]]) -> float:
    conflicts = 0
    for node, neighbors in graph.items():
        for n in neighbors:
            if chromosome[node] == chromosome[n]:
                conflicts += 1
    conflicts //= 2
    return 1.0 / (1.0 + conflicts)


def selection(population: list[list[int]], graph: dict[int, list[int]], k: int = 3) -> list[int]:
    selected = random.sample(population, k)
    selected.sort(key=lambda x: fitness(x, graph), reverse=True)
    return selected[0]


def crossover(parent1: list[int], parent2: list[int]) -> tuple[list[int], list[int]]:
    if len(parent1) > 2:
        point = random.randint(1, len(parent1) - 2)
        child1 = parent1[:point] + parent2[point:]
        child2 = parent2[:point] + parent1[point:]
        return child1, child2
    return parent1[:], parent2[:]


def mutate(chromosome: list[int], mutation_rate: float, num_colors: int) -> list[int]:
    for i in range(len(chromosome)):
        if random.random() < mutation_rate:
            old_color = chromosome[i]
            new_color = random.randint(0, num_colors - 1)
            while new_color == old_color:
                new_color = random.randint(0, num_colors - 1)
            chromosome[i] = new_color
    return chromosome


def map_coloring_ga(
    graph: dict[int, list[int]],
    num_colors: int,
    pop_size: int = 100,
    generations: int = 500,
    crossover_rate: float = 0.8,
    mutation_rate: float = 0.01,
    seed: int | None = None,
) -> tuple[list[int] | None, float]:
    if seed is not None:
        random.seed(seed)

    num_genes = len(graph)
    population = initial_population(pop_size, num_genes, num_colors)
    best_solution: list[int] | None = None
    best_fitness = 0.0

    for _ in range(generations):
        fitness_values = [fitness(ch, graph) for ch in population]
        best_index = fitness_values.index(max(fitness_values))
        current_best = population[best_index]
        current_best_fit = fitness_values[best_index]

        if current_best_fit > best_fitness:
            best_solution = current_best[:]
            best_fitness = current_best_fit

        if best_fitness == 1.0:
            break

        population.sort(key=lambda ch: fitness(ch, graph), reverse=True)
        new_population = population[:2]

        while len(new_population) < pop_size:
            parent1 = selection(population, graph)
            parent2 = selection(population, graph)

            if random.random() < crossover_rate:
                child1, child2 = crossover(parent1, parent2)
            else:
                child1, child2 = parent1[:], parent2[:]

            mutate(child1, mutation_rate, num_colors)
            mutate(child2, mutation_rate, num_colors)
            new_population.extend([child1, child2])

        population = new_population[:pop_size]

    return best_solution, best_fitness


def run_graph_coloring_ga(seed: int | None = None) -> dict:
    graph = {
        0: [1, 2],
        1: [0, 2, 3],
        2: [0, 1, 4, 3],
        3: [1, 4, 2],
        4: [2, 3],
    }
    solution, fit = map_coloring_ga(graph, num_colors=3, pop_size=50, generations=100, seed=seed)
    return {"graph": graph, "solution": solution, "fitness": fit}
